#include "env.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pebblisp.h"
#include "web.h"
#include "threads.h"
#include "plfunc.h"

struct symFunc buildFuncSym(const char* symbol, Object (* func)(Object*, int, struct Environment*), const char* help,
                            const char* const* tests, size_t testLength);

struct Environment* globalEnv;

struct Environment* global()
{
    return globalEnv;
}

void setGlobal(struct Environment* env)
{
    globalEnv = env;
}

struct Dictionary {
    int structCount;
    int structCapacity;
    struct StructDef* structDefs;
} dictionary;

/// For any instances (e.g. segfault recovery) where defaultEnv() may be called
/// multiple times.
int helpInitialized = 0;

#ifdef STANDALONE
struct helpText {
    const char* symbol;
    const char* help;
    const char* const* tests;
    size_t testCount;
};
int currentHelp = 0;
struct helpText helpTexts[100];
#endif

struct symFunc {
    const char* sym;
    Object (* func)(Object*, int, struct Environment*);
};

#ifdef STANDALONE
#define pf(_func) buildFuncSym(_func ## Symbol, &(_func), _func ## Doc, _func ## Tests, array_length(_func ## Tests))
#else
#define pf(_func) buildFuncSym(_func ## Symbol, &(_func))
#endif

struct StrippedObject* fetch(const char* name, struct Environment* env)
{
    while (env) {
        struct StrippedObject* object = getFromTable(&env->table, name);
        if (object) {
            return object;
        }
        env = env->outer;
    }
    return NULL;
}

Object fetchFromEnvironment(const char* name, struct Environment* env)
{
    struct StrippedObject* object = fetch(name, env);
    if (object) {
        Object o = deStrip(*object);
        return cloneObject(o);
    }
    throw(DID_NOT_FIND_SYMBOL, "%s", name);
}

void addToEnv(struct Environment* env, const char* name, const Object obj)
{
    struct StrippedObject* existing = fetch(name, env);
    if (existing) {
        Object o = deStrip(*existing);
        cleanObject(&o);
        o = cloneObject(obj);
        existing->type = o.type;
        existing->data = o.data;
        return;
    }
    char* symbol = malloc(sizeof(char) * (strlen(name) + 1));
    strcpy(symbol, name);
    addToTable(&env->table, symbol, cloneObject(obj));
}

/**
 * Build the local Environment for a given lambda
 * @param params List of symbol objects with the names of lambda params
 * @param paramCount Length of params list
 * @param arguments Arguments being passed into the lambda where ->forward is the next arg
 * @param outer The parent scope
 * @return The constructed environment
 */
struct Environment envForLambda(const Object* params, const Object* arguments, int paramCount,
                                struct Environment* outer)
{
    if (outer) {
        outer->refs += 1;
    }

    struct Environment env = {
        .outer = NULL,
        .table = buildTable(paramCount * 2),
        .refs = 1,
    };

    const Object* argument = arguments;
    for (int i = 0; i < paramCount; i++) {
        const char* paramName = itemAt(params, i)->string;
        // Eval the `argument` list
        Object newEnvObj = argument ? eval(argument, outer) : errorWithContext(NOT_ENOUGH_ARGUMENTS, paramName);
        addToEnv(&env, paramName, newEnvObj); // Could use eval_forms?
        cleanObject(&newEnvObj);
        argument = argument ? argument->forward : NULL;
    }
    env.outer = outer;

    return env;
}

void deleteEnv(struct Environment* e)
{
    e->refs -= 1;
    if (e->refs) {
        return;
    }

    if (e->outer) {
        deleteEnv(e->outer);
    }

    deleteTable(&e->table);
}


struct Environment defaultEnv()
{
#ifndef STANDALONE
    int helpInitialized = 0;
#endif
    if (!helpInitialized) {
        dictionary = (struct Dictionary) {
            .structCount = 0,
            .structCapacity = 8,
            .structDefs = malloc(sizeof(struct StructDef) * 8),
        };
    }

    struct Environment e = {
        .table = buildTable(128),
        .outer = NULL,
        .refs = 1,
    };

    struct symFunc symFuncs[] = {
        pf(def),
        pf(add),
        pf(sub),
        pf(mul),
        pf(dvi),
        pf(mod),
        pf(equ),
        pf(greaterThan),
        pf(lessThan),
        pf(and),
        pf(or),
        pf(catObjects),
        pf(filter),
        pf(len),
        pf(append),
        pf(prepend),
        pf(reduce),
        pf(mapO),
        pf(at),
#ifndef PBL_PLATFORM_APLITE
        pf(rest),
        pf(charAt),
        pf(chars),
        pf(matches),
        pf(slen),
        pf(substring),
        pf(isNum),
        pf(isList),
        pf(isString),
        pf(isErr),
        pf(charVal),
        pf(parseEvalO),
#endif
        pf(structAccess),
        pf(getTime),
#ifndef LOW_MEM
        pf(reverse),
#endif
#ifdef WEBSERVER
        pf(addGetRoute),
        pf(addPostRoute),
        pf(startServer),
#endif
#ifdef STANDALONE
        pf(segfault),
        pf(print),
        pf(numToChar),
        pf(printEnvO),
        pf(systemCall),
        pf(cd),
        pf(cwd),
        pf(takeInput),
        pf(readFileToObject),
        pf(getEnvVar),
        pf(async),
        pf(await),
        pf(help)
#endif
    };

    unsigned i;
    for (i = 0; i < sizeof(symFuncs) / sizeof(symFuncs[0]); i++) {
        addFunc(symFuncs[i].sym, symFuncs[i].func, &e);
    }

    helpInitialized = 1;
    return e;
}

void printEnv(struct Environment* env, int printPointers)
{
    if (!env) {
        printf("NULL env\n");
        return;
    }
    printf("env->capacity = %lu\n", env->table.capacity);
    printf("env->count = %lu\n", env->table.count);

    for (int i = 0; i < env->table.capacity; i++) {
        printf("[0m");
        if (env->table.elements[i].symbol == NULL) {
            continue;
        }
        printf("%s: ", env->table.elements[i].symbol);
        printf("[0m");
        if (env->table.elements[i].object.type == TYPE_STRING) {
            printf("\"");
        }
        if (env->table.elements[i].object.type == TYPE_FUNC && !printPointers) {
            printf("Native");
        } else {
            size_t length;
            // SAFETY: Casting to Object is fine because stringObj() doesn't touch .forward
            char* s = stringObj((Object *) &env->table.elements[i].object, &length);
            printColored(s);
            if (env->table.elements[i].object.type == TYPE_STRING) {
                printf("\"");
            }
            free(s);
        }
        if (printPointers) {
            printf(" @ %p", env->table.elements[i].symbol);
        }
        printf("\n");
    }
}

void addFunc(const char* name,
             Object (* func)(Object*, int, struct Environment*),
             struct Environment* env)
{
    Object o = newObject(TYPE_FUNC);
    o.func = func;
    addToEnv(env, name, o);
}

void shredDictionary()
{
    for (int i = 0; i < dictionary.structCount; i++) {
        free(dictionary.structDefs[i].name);
        for (int j = 0; j < dictionary.structDefs[i].fieldCount; j++) {
            free(dictionary.structDefs[i].names[j]);
        }
        free(dictionary.structDefs[i].names);
    }
    free(dictionary.structDefs);
}

#ifdef STANDALONE

struct symFunc buildFuncSym(const char* symbol, Object (* func)(Object*, int, struct Environment*), const char* help,
                            const char* const* tests, size_t testLength)
{
    if (!helpInitialized) {
        helpTexts[currentHelp].help = help;
        helpTexts[currentHelp].symbol = symbol;
        helpTexts[currentHelp].tests = tests;
        helpTexts[currentHelp].testCount = testLength;
        currentHelp += 1;
    }

    return (struct symFunc) {
        .func = func,
        .sym = symbol,
    };
}

#else
struct symFunc buildFuncSym(const char* symbol, Object (* func)(Object*, int, struct Environment*))
{
    return (struct symFunc) {
        .func = func,
        .sym = symbol,
    };
}
#endif

#ifdef STANDALONE
const int black = 30;
const int red = 31;
const int green = 32;
const int yellow = 33;
const int cyan = 34;
const int purple = 35;
const int teal = 36;
const int white = 37;
const int colors[] = {
    white, green, cyan, yellow, purple, red
};
const int colorCount = array_length(colors);

int getColor(int depth)
{
    depth = depth >= 0 ? depth : depth * -1;
    return colors[depth % colorCount];
}

void printColored(const char* code)
{
    int c = 0;
    int depth = 0;
    int isQuote = 0;
    printf("[%dm", getColor(depth));
    while (code[c]) {
        if ((code[c] == '(' || code[c] == '{') && !isQuote) {
            depth += 1;
            printf("[%dm", getColor(depth));
        } else if ((code[c] == ')' || code[c] == '}') && !isQuote) {
            depth -= 1;
            printf("%c[%dm", code[c], getColor(depth));
            c++;
            continue;
        } else if (code[c] == '"') {
            isQuote = !isQuote;
            if (isQuote) {
                printf("[%dm\"", teal);
            } else {
                printf("\"[%dm", getColor(depth));
            }
            c++;
            continue;
        } else if (code[c] == '=' && code[c + 1] == '>' && !isQuote) {
            printf("[1m=>[0;%dm", getColor(depth));
            c += 2;
            continue;
        }
        printf("%c", code[c]);
        c++;
    }
}

Object segfault(Object* params, int length, struct Environment* env)
{
    int* p = NULL;
    return numberObject(*p);
}

Object help(Object* params, int length, struct Environment* env)
{
    const char* symbol;
    if (!length || !params[0].string || params[0].string[0] == '\0') {
        symbol = "?";
    } else {
        symbol = params[0].string;
    }
    for (int i = 0; i < currentHelp; i++) {
        struct helpText h = helpTexts[i];
        if (strcmp(symbol, h.symbol) == 0) {
            Object text = withLen(1024, TYPE_STRING);
            char* textCursor = text.string;
            textCursor += sprintf(textCursor, "%s", h.help);
            for (int ti = 0; ti < h.testCount; ti += 2) {
                const char* test = h.tests[ti];
                const char* expected = h.tests[ti + 1];
                if (test[0] == ';') {
                    textCursor += sprintf(textCursor, "\n  %s => %s", test + 1, expected);
                    continue;
                }
                textCursor += sprintf(textCursor, "\n  %s => ", test);
                if (expected[0] == '\0') {
                    textCursor += sprintf(textCursor, "<empty string>");
                } else {
                    textCursor += sprintf(textCursor, "%s", expected);
                }
            }
            return text;
        }
    }

    for (int i = 0; i < dictionary.structCount; i++) {
        if (strcmp(dictionary.structDefs[i].name, symbol) == 0) {
            char structDef[128] = {'{', ' ', '\0'};
            for(int field = 0; field < dictionary.structDefs[i].fieldCount; field++) {
                strcat(structDef, dictionary.structDefs[i].names[field]);
                strcat(structDef, " ");
            }
            strcat(structDef, "}");
            return nullTerminated(structDef);
        }
    }

    struct StrippedObject* object = fetch(symbol, env);
    if (object) {
        size_t len = 0;
        // SAFETY: Casting to Object is fine because stringObj() doesn't touch .forward
        char* string = stringObj((Object*) object, &len);
        Object text = stringFromSlice(string, len);
        free(string);
        return text;
    }

    return nullTerminated("Help not found!");
}

// Returns number of failures, or -1 if the requested specificTest is invalid
int runTests(int detailed, int specificTest)
{
    int isSpecific = specificTest != -1;
    if (specificTest >= currentHelp) {
        return -1;
    }

    if (!isSpecific) {
        printf("[35;1m::NATIVE TESTS::[0m\n");
    }
    int failureCount = 0;
    int passCount = 0;
    int start = isSpecific ? specificTest : 0;
    int end = isSpecific ? specificTest + 1 : currentHelp;
    for (int hi = start; hi < end; hi++) {
        struct helpText h = helpTexts[hi];
        if (h.tests) {
            if (detailed && h.testCount > 0) {
                if (isSpecific) {
                    printf("[0m`%s` ", h.symbol);
                } else {
                    printf("[0m `%s` ", h.symbol);
                }
            }
            for (int ti = 0; ti < h.testCount; ti += 2) {
                const char* test = h.tests[ti];
                const char* expected = h.tests[ti + 1];
                if (test[0] == ';') {
                    continue;
                }
                struct Environment env = defaultEnv();
                Object o = parseEval(test, &env);
                size_t length;
                char* result = stringObj(&o, &length);
                cleanObject(&o);
                int expectedLen = 0;
                while (expected[expectedLen] && expected[expectedLen] != ';') {
                    expectedLen++;
                }
                if (strncmp(result, expected, expectedLen) != 0) {
                    failureCount++;
                    printf("Test failed!\n");
                    printf("%s\n", test);
                    printf("Expected '%s' but received '%s'\n", expected, result);
                } else {
                    if (detailed) {
                        printf("[32;1m✓");
                    }
                    passCount++;
                }
                free(result);
                deleteEnv(&env);
            }
            if (detailed && h.testCount > 0) {
                printf("\n");
            }
        }
    }
    if (passCount > 0) {
        printf("[32;1m");
    }
    if (!isSpecific) {
        printf("%d tests passed![0m\n", passCount);
    }
    if (failureCount > 0) {
        printf("[31m");
    }
    if (!isSpecific || failureCount != 0) {
        printf("%d tests failed![0m\n", failureCount);
    }
    // fprintf(stderr, "TOTAL ALLOCATIONS: %d\n", getAllocations());
    // fprintf(stderr, "TOTAL BYTES: %zu\n", getBytes());
    return failureCount;
}

#endif

int getStructIndex(const char* name)
{
    for (int i = 0; i < dictionary.structCount; i++) {
        if (strcmp(name, dictionary.structDefs[i].name) == 0) {
            return i;
        }
    }
    return -1;
}

struct StructDef* getStructAt(int i)
{
    return &dictionary.structDefs[i];
}

void addStructDef(struct StructDef def)
{
    dictionary.structDefs[dictionary.structCount] = def;
    dictionary.structCount += 1;
    if (dictionary.structCount == dictionary.structCapacity) {
        struct StructDef* prev = dictionary.structDefs;
        int prevCapacity = dictionary.structCapacity;
        dictionary.structCapacity *= 2;
        dictionary.structDefs = malloc(sizeof(struct StructDef) * dictionary.structCapacity);
        for (int i = 0; i < prevCapacity; i++) {
            dictionary.structDefs[i] = prev[i];
        }
        free(prev);
    }

}
